from typing import Any
from uuid import UUID
from sqlalchemy import Dialect, types


class char_UUID(types.TypeDecorator):
    """
    Mysql has no native UUID type, therefore we store it as a CHAR with length 32.
    """

    impl = types.CHAR(32)
    # we user char(32) as types.Uuid excepts a UUID on filter... You cant make partial uuids.

    cache_ok = True

    def process_bind_param(self, value: Any | None, dialect: Dialect) -> Any:
        if not value:
            return value

        if isinstance(value, UUID):
            return value.hex

        return str(value).replace("-", "")

    def process_result_value(self, value: str | None, dialect: Dialect) -> Any | None:
        if value:
            return str(UUID(hex=value))
        return value



from sqlalchemy import BigInteger, Table, MetaData, Column, Integer, String

LOGGING_CONFIG = {
    "version": 1,
    "disable_existing_loggers": True,
    "formatters": {
        "standard": {"format": "%(asctime)s - %(levelname)-8s - %(message)s"},
    },
    "handlers": {
        "default": {
            "level": "INFO",
            "formatter": "standard",
            "class": "logging.StreamHandler",
            "stream": "ext://sys.stdout",  # Default is stderr
        },
    },
    "loggers": {
        "": {  # root logger
            "handlers": ["default"],
            "level": "WARNING",
            "propagate": False,
        },
        "sqlalchemy.engine": {
            "handlers": ["default"],
            "level": "WARNING",
            "propagate": False,
        },
        "__main__": {  # if __name__ == '__main__'
            "handlers": ["default"],
            "level": "DEBUG",
            "propagate": False,
        },
        "alembic": {"handlers": ["default"], "level": "INFO", "propagate": False},
        "SqlMigrate": {"handlers": ["default"], "level": "INFO", "propagate": False},
    },
}

schema_migrations = "schema_migrations"

schema_migrations_table = Table(
    schema_migrations,
    MetaData(),
    Column("id", Integer, primary_key=True),
    Column("version_num", BigInteger),
    Column("name", String(256)),
    Column("duration", Integer),
    Column("migrated", Integer),
)



import datetime
import re
import socket
from collections import OrderedDict
from logging.handlers import SysLogHandler, SYSLOG_UDP_PORT
from syslog import LOG_USER
from typing import Tuple

NILVALUE = "-"

# As defined in RFC5424 Section 7
STRUCTURED_DATA_IDs = ("timeQuality", "origin", "meta")

# Version of the protocol we support
SYSLOG_VERSION = 1


class FullSysLogHandler(SysLogHandler):
    def __init__(
        self,
        appname: str,
        address: Tuple[str, int] = ("localhost", SYSLOG_UDP_PORT),
        facility: int = LOG_USER,
        socktype: socket.SocketKind = socket.SOCK_DGRAM,
        hostname: str = None,
        procid: int = None,
        msgid: int = None,
        structured_data: OrderedDict = None,
        enterprise_id: int = None,
    ):
        if structured_data is None:
            structured_data = OrderedDict()

        super().__init__(address, facility, socktype)

        self.hostname, self.appname, self.procid = hostname, appname, procid
        self.msgid, self.structured_data, self.enterprise_id = (
            msgid,
            structured_data,
            enterprise_id,
        )

        if self.hostname is None:
            self.hostname = socket.gethostname()

        if self.procid is None:
            self.procid = NILVALUE

        if self.msgid is None:
            self.msgid = NILVALUE

    """An RFC 5425-complaint Syslog Handler for the python logging framework"""

    def get_hostname(self, record):
        return getattr(record, "hostname", self.hostname)

    def get_appname(self, record):
        return getattr(record, "appname", self.appname)

    def get_procid(self, record):
        # returns the record.name field instead of procid; which is the python logger id; e.g. web.app.run
        return getattr(record, "name", self.procid)

    def get_msgid(self, record):
        return getattr(record, "msgid", self.msgid)

    def get_enterprise_id(self, record):
        return getattr(record, "enterprise_id", self.enterprise_id)

    def get_structured_data(self, record):
        structured_data = OrderedDict()
        structured_data.update(self.structured_data)
        structured_data.update(getattr(record, "structured_data", {}))
        return structured_data

    def emit(self, record):
        """
        Emit a record.

        The record is formatted, and then sent to the syslog server. If
        exception information is present, it is NOT sent to the server.
        """
        # Prepare the message.

        # First the priority:
        pri = "<%d>%d" % (
            self.encodePriority(self.facility, self.mapPriority(record.levelname)),
            SYSLOG_VERSION,
        )

        # Let's map some extra stuff, and clean it up to max values
        hostname = (
            self.get_hostname(record).encode("ascii", "replace")[:255].decode("utf-8")
        )
        appname = (
            self.get_appname(record).encode("ascii", "replace")[:48].decode("utf-8")
        )
        procid = (
            self.get_procid(record).encode("ascii", "replace")[:128].decode("utf-8")
        )
        msgid = self.get_msgid(record).encode("ascii", "replace")[:32].decode("utf-8")
        enterprise_id = self.get_enterprise_id(record)

        structured_data = self.get_structured_data(record)

        # Clean up the structured data
        for key, value in list(structured_data.items()):
            # we ask for the list, so we can change the keys

            if isinstance(value, dict):
                value = value.items()
            newvals = []
            for itemkey, itemvalue in value:
                itemkey = (
                    itemkey.replace('"', "")
                    .replace(" ", "")
                    .replace("]", "")
                    .replace("=", "")[:32]
                    # .encode("ascii", "replace")
                )
                itemvalue = (
                    itemvalue.replace("\\", "\\\\")
                    .replace('"', '\\"')
                    .replace("]", "\\]")
                    # .encode("utf8", "replace")
                )
                newvals.append('%s="%s"' % (itemkey, itemvalue))

            structured_data[key] = " ".join(newvals)

            if (
                key not in STRUCTURED_DATA_IDs
                and enterprise_id is not None
                and "@" not in key
            ):
                newkey = "%s@%s" % (key, enterprise_id)
            else:
                newkey = key
                newkey = (
                    newkey.replace('"', "")
                    .replace(" ", "")
                    .replace("]", "")
                    .replace("=", "")[:32]  # .encode("ascii", "replace")
                )

            if newkey != key:
                structured_data[newkey] = structured_data[key]
                del structured_data[key]

        msg = self.format(record)

        # parse out ansi color tags..
        ansi_escape = re.compile(r"\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])")
        msg = ansi_escape.sub("", msg)

        ts = datetime.datetime.fromtimestamp(record.created)
        # NEED TO DEAL WITH TIME ZONES
        timestamp = ts.isoformat()

        header = f"{pri} {timestamp} {hostname} {appname} {procid} {msgid}"

        if structured_data:
            sd = "".join(
                "[%s]" % (" ".join((key, value)))
                for key, value in structured_data.items()
            )
        else:
            sd = NILVALUE

        msg = " ".join((header, sd, msg, "\000")).encode("utf-8")

        # This section copied from logging.SyslogHandler
        # noinspection PyBroadException
        try:  # pragma: no cover
            if self.unixsocket:
                try:
                    self.socket.send(msg)
                except OSError:
                    self.socket.close()
                    self._connect_unixsocket(self.address)
                    self.socket.send(msg)
            elif self.socktype == socket.SOCK_DGRAM:
                self.socket.sendto(msg, self.address)
            else:
                self.socket.sendall(msg)
        except (KeyboardInterrupt, SystemExit):  # pragma: no cover
            raise
        except Exception:  # pragma: no cover
            self.handleError(record)



from typing import Any, Optional, Union, Dict, TextIO, Sequence

from alembic import util
from alembic.config import Config
from alembic.runtime.environment import (
    EnvironmentContext,
    CompareType,
    CompareServerDefault,
    RenderItemFn,
    ProcessRevisionDirectiveFn,
    IncludeObjectFn,
    IncludeNameFn,
    OnVersionApplyFn,
)
from alembic.script import ScriptDirectory
from sqlalchemy import Connection, URL, MetaData

from ...fractured_alembic.errors.context import ContextNotConfigured
from ...fractured_alembic.runtime.migration import FracturedMigrationContext


class SqlEnvironmentContext(EnvironmentContext):
    def __init__(self, config: Config, script: ScriptDirectory, **kw: Any):
        super().__init__(config, script, **kw)

    def configure(
        self,
        connection: Optional[Connection] = None,
        url: Optional[Union[str, URL]] = None,
        dialect_name: Optional[str] = None,
        dialect_opts: Optional[Dict[str, Any]] = None,
        transactional_ddl: Optional[bool] = None,
        transaction_per_migration: bool = False,
        output_buffer: Optional[TextIO] = None,
        starting_rev: Optional[str] = None,
        tag: Optional[str] = None,
        template_args: Optional[Dict[str, Any]] = None,
        render_as_batch: bool = False,
        target_metadata: Union[MetaData, Sequence[MetaData], None] = None,
        include_name: Optional[IncludeNameFn] = None,
        include_object: Optional[IncludeObjectFn] = None,
        include_schemas: bool = False,
        process_revision_directives: Optional[ProcessRevisionDirectiveFn] = None,
        compare_type: Union[bool, CompareType] = True,
        compare_server_default: Union[bool, CompareServerDefault] = False,
        render_item: Optional[RenderItemFn] = None,
        literal_binds: bool = False,
        upgrade_token: str = "upgrades",
        downgrade_token: str = "downgrades",
        alembic_module_prefix: str = "op.",
        sqlalchemy_module_prefix: str = "sa.",
        user_module_prefix: Optional[str] = None,
        on_version_apply: Optional[OnVersionApplyFn] = None,
        **kw: Any,
    ) -> None:
        opts = self.context_opts
        if transactional_ddl is not None:
            opts["transactional_ddl"] = transactional_ddl
        if output_buffer is not None:
            opts["output_buffer"] = output_buffer
        elif self.config.output_buffer is not None:
            opts["output_buffer"] = self.config.output_buffer
        if starting_rev:
            opts["starting_rev"] = starting_rev
        if tag:
            opts["tag"] = tag
        if template_args and "template_args" in opts:
            opts["template_args"].update(template_args)
        opts["transaction_per_migration"] = transaction_per_migration
        opts["target_metadata"] = target_metadata
        opts["include_name"] = include_name
        opts["include_object"] = include_object
        opts["include_schemas"] = include_schemas
        opts["render_as_batch"] = render_as_batch
        opts["upgrade_token"] = upgrade_token
        opts["downgrade_token"] = downgrade_token
        opts["sqlalchemy_module_prefix"] = sqlalchemy_module_prefix
        opts["alembic_module_prefix"] = alembic_module_prefix
        opts["user_module_prefix"] = user_module_prefix
        opts["literal_binds"] = literal_binds
        opts["process_revision_directives"] = process_revision_directives
        opts["on_version_apply"] = util.to_tuple(on_version_apply, default=())

        if render_item is not None:
            opts["render_item"] = render_item
        opts["compare_type"] = compare_type
        if compare_server_default is not None:
            opts["compare_server_default"] = compare_server_default
        opts["script"] = self.script

        opts.update(kw)

        self._migration_context = FracturedMigrationContext.configure(
            connection=connection,
            url=url,
            dialect_name=dialect_name,
            environment_context=self,
            dialect_opts=dialect_opts,
            opts=opts,
        )

    def get_context(self) -> FracturedMigrationContext:
        if self._migration_context is None:
            raise ContextNotConfigured("No context has been configured yet.")
        return self._migration_context

    def __repr__(self):
        return f"<< {self.__class__.__name__ } >>"
